# 連立一次ディオファントス方程式 	lde(M,b)
# 整数行列の単因子標準形			elmdiv(M)
# ver.1.0  (2020-12-01)
# ver.1.0a  (2020-12-05)
# begin with:
# import mlde
# coded by Kenji Arisawa
# email: arisawa@aichi-u.ac.jp

# 文献
# https://docs.sympy.org/latest/modules/
# https://docs.python.org/3/

# hand translation from lde.mac

# start python3 and then execute:
"""-
import mlde
from mlde import *
-"""

# comment to developers
# avoiding to restart python:
lds="importlib.reload(mlde);from mlde import *"	# _L is private to the module
"""-
exec(lds)
# NG:
# def ld():
#  importlib.reload(mlde)
# we need to avoid 'lde.py', otherwise name confliction!
-"""

def ld(): print("exec(lds) # do this one")

import importlib	# used only in exec(_L). you may comment out
# look https://docs.python.org/3.6/library/importlib.html#importlib.reload for importlib.reload

from sympy import Matrix,eye,symbols,expand,det,latex
from random import randint

debug= True
verbose=True


#
#	some check points in translation
#
"""-
grep -n ',[ ]*$' lde.py
-"""

#
#	some notes on Sympy
#
"""-
(1) Vector
vectors in Sympy are special matrix.
they are either single col or single row matrix.
(2) Matrix in Sympy is very flexible data type
(3) Slice on matrix M is great! very powerful!
-"""


#
#	some tools
#

def error(*t):
  raise Exception(*t)

# sympy vector is a special matrix
def vlen(v): # vector length
  s = v.shape
  if min(s) != 1: raise Exception("not a vector")
  return max(s)

# 'Matrix' of sympy is a pain for me
# You might think 'matrix' may be better, 
# however I am afraid the name may conflict with other modules such as numpy.
def Mat(*L):
  return Matrix([*L])

# handy test if x is zero or zero matrix
# look https://docs.python.org/ja/3/library/functions.html#getattr for getattr
def is_zero(x):
  if x == 0: return True
  if hasattr(x,"is_zero_matrix"): return x.is_zero_matrix
  return False

# it may be wise to have sublist() of Maxima
def sublist(L,f):
  return list(filter(f,L))  # need list(...) for Python3

"""-
L=[5,-3,2,-5,3,-7]
sublist(L,lambda x: x>0)	# [5, 2, 3]
-"""  


# tab_gen(M) returns new tablet [S,A,T]
# tab_gen2(M,b) is for linear diophantine equations 
def tab_gen(M):
  nrow,ncol = M.shape
  return [eye(nrow),M.copy(),eye(ncol)]
def tab_gen2(M,b):
  nrow,ncol = M.shape
  if nrow != len(b): return False
  return [Mat(b).transpose(),M.copy(),eye(ncol)]
	
# Maxima div(x,d) is x//d, while sympy div(x,d) is for polynomial division

# multiply c to row n of M
def mat_mulrow(M,n,c):
  M[n,0:]=c*M[n,0:]
  return M	# original M is untouched

#  multiply c to col n of M
def mat_mulcol(M,n,c):
  M[0:,n]=c*M[0:,n]
  return M	# original M is untouched
	
# mat_min(M): a function that gets minimum elements in matrix M
# no need in Python. min(M) works fine
  
#  mat_pivot(M) returns minimum (>0) index in abs(M),
#  which is something like a pivot for getting elementary divisor decomposition
def mat_pivot(M): # return tuple of the point
  X=abs(M)
  L=usort(X)
  m=L[0]     # minimum value in X
  if m == 0:
    if len(L) == 1: return None
    m=L[1]
  return mat_index(X,m)
"""- 
M = Mat([-4,3],[-2,5],[3,5])
mat_pivot(M) # → (1,0)

M=Mat(
[1, 0, 0],
[0, 3, 0],
[0, 0, 0])
mat_pivot(mat_subm(M,1))	# → (0,0)
-"""

# returns one of index (i,j) such that M[i,j]==a if exist, otherwise raise error
# we don't need a filter for 'a'
def mat_index(M,a): 
	return divmod(M[0:].index(a),M.shape[1])
	# M[0:] is a list that produced M

# unique sort, that is, sort without duplication
# X is list or Matrix
# no need mat_usort(). usort() works for both list and matrix
def usort(X): # return a list
  return sorted(list(set(X)))

# submatrix of M that begins with M[k,k] to the end
def mat_subm(M,k):
  return M[k:,k:,]
  
def mat_swaprow(M,i,j):
  M[i,0:],M[j,0:] = M[j,0:],M[i,0:]
  return M
  
def mat_swapcol(M,i,j):
  M[0:,i],M[0:,j] = M[0:,j],M[0:,i]
  return M
  
"""-
M = Mat([-4,-2,5],[3,5,7])
mat_swaprow(M,0,1) 	# → Mat([3,5,7],[-4,-2,5]), # M is changed 
mat_swapcol(M,0,1)  # → Mat([-2,-4,5],[5,3,7]), # M is kept 
-"""

# mat_addrow(M,i,L)
# mat_addcol(M,j,L)
def mat_addrow(M,i,L):	# L is row vector
  A = M.copy()
  A[i,0:] = A[i,0:]+L
  return A
  
def mat_addcol(M,j,L):	# L is col vector
  A = M.copy()
  A[0:,j] = A[0:,j]+L
  return A
  
"""-
L = [1,2,3]
M = Mat([-4,-2,5],[3,5,7])
mat_addrow(M,0,L)  # → Mat([-3,0,8],[3,5,7])
mat_addcol(M,1,[6,1])  # → Mat([-4,4,5],[3,6,7])
-"""

# we don't need mat_getrow(M,i) and mat_getcol(M,j)
# instead use M.row(i) and M.col(j). they are vector, not lists


# mat_getdia(M)	# diagonal elements in M
def mat_getdia(M):
  return list(M.diagonal())
"""-
M = Mat([-4,-2,5],[3,5,7])
mat_getdia(M)	# → [-4,5]
-"""


# diag() in sympy is clumsy! I want to have diagl(L) for a list L
# instead the function below is much more powerful
def genmatrix(f,nrow,ncol):
  M = zeros(nrow,ncol)
  for i in range(0,nrow):
    for j in range(0,ncol):
      M[i,j] = f(i,j)
  return M

"""-
M = Mat([1,2,3],[4,5,6]) 
nrow,ncol = M.shape
genmatrix(lambda i,j: int(i==j)*M[i,j],nrow,ncol)  # → Matrix([[1, 0, 0],[0, 5, 0]])
-"""

# sympy has a deprecated module "mod". it is wise not to use the name 'mod'
# the xmod(p,d) below works for matrix p
# note that xmod(0,0) returns 0
def xmod(p,d):
  if is_zero(p): return 0
  return p%d
  

# tab_fix(B,k)		# fix B so that B[2][k,k]>0 */
def tab_fix(B,k):
  S,A,T = B[0],B[1],B[2]
  if A[k,k] < 0:
    A = mat_mulrow(A,k,-1)
    S = mat_mulrow(S,k,-1)
  return([S,A,T])

# check if M[i,j]==0 (i>=k, j>=k)
def chk_zero(M,k):
  return is_zero(mat_subm(M,k))

# check if M[i,k]==0 (i != k, i=1,2,...)
def chk_col(M,k):
  v = M.col(k); v[k] = 0
  return is_zero(v)

# check if M[k,j]=0 (j != k, j=1,2,...)
def chk_row(M,k):
  v = M.row(k); v[k] = 0
  return is_zero(v)

"""-
chk_col(Mat([2,4,1],[0,3,2]),0) 	# → True
chk_row(Mat([2,0,0],[5,3,2]),0) 	# → true
-"""

# check if we may increase k
def chk_mod(M,k):
  M1 = mat_subm(M,k)
  if is_zero(M1): return True
  if M[k,k] == 0: return False
  M1 = M1 % M[k,k]
  if is_zero(M1): return True
  return False

# final check
def chk_final(B,M):
  S,A,T = B[0],B[1],B[2]
  L = mat_getdia(A)
  if len(sublist(L,lambda x: x<0)) >0: return 1
  if A.is_diagonal() == False: return 2
  m = min(A.shape)
  for n in range(0,m-1):
    if xmod(L[n+1],L[n]) != 0: return 3
  if S*M*T != A: return 4
  return(0)


#
# some debugging tools
#
def dbg(B,M):
  if debug and B[1]*M*B[3] != B[2]: error("# ERROR",B,M)
  return True

def dprint(str,k,B):
  if verbose: print(str,k+1,B)

debug=True  # the default is True

# move the minimum element(>0) in abs(mat_subm(A,k)) to A[k,k]
def step1(B,k):
  B = tab_fix(B,k)
  S,A,T = B[0],B[1],B[2]
  L = mat_pivot(mat_subm(A,k))	# no need false check for L
  if L == None: return False
  i,j=L
  if i != 0 : A=mat_swaprow(A,k,k+i); S=mat_swaprow(S,k,k+i)
  if j != 0 : A=mat_swapcol(A,k,k+j); T=mat_swapcol(T,k,k+j)
  return [S,A,T]


# make absolute values A[k+1,k],A[k+2,k],... , S[k+1,k],S[k+2,k],... smaller using modulo
def step2(B,k):
  B = tab_fix(B,k)
  S,A,T = B[0],B[1],B[2]
  L = A[k,0:]
  m = L[k]
  L1 = S[k,0:]
  for n in range(k+1,A.shape[0]):
    q = A[n,k]//m
    A[n,0:] = A[n,0:]-q*L
    S[n,0:] = S[n,0:]-q*L1
  return [S,A,T]


# make absolute values A[k,k+1],A[k,k+2],... , T[k,k+1],T[k,k+2],... smaller using modulo
def step3(B,k):
  B = tab_fix(B,k)
  S,A,T = B[0],B[1],B[2]
  L0 = A[k,0:]
  m = L0[k] 	# m is positive
  L2 = A[0:, k]
  L3 = T[0:, k]
  for n in range(k+1,vlen(L0)):
    q = L0[n]//m
    A = mat_addcol(A,n,-q*L2)   
    T = mat_addcol(T,n,-q*L3)
  return [S,A,T]

"""-
B = [Mat([2,1],[-5,-2]),Mat([1,-5,17],[0,14,-39]),Mat([1,1,0],[0,0,0],[0,0,1])];
step3(B,0) 
  # → [Mat([2,1],[-5,-2]),Mat([1,0,0],[0,14,-39]),
  # Mat([1,6,-17],[0,0,0],[0,0,1])], OK 
-"""



# fix:  B is assumed that both chk_col(B[1],k) and chk_row(B[1],k) are true.
#  then we can make B[1][k,k] smaller
def step4(B,k):
  B = tab_fix(B,k)
  S,A,T = B[0],B[1],B[2]
  L = mat_pivot(xmod(mat_subm(A,k),A[k,k]))
  if L == None : return False
  q = A[L[0],L[1]]//A[k,k]	# quotient
  L = (L[0]+k,L[1]+k)
  A = mat_addrow(A,L[0],q*A[k,0:])
  S = mat_addrow(S,L[0],q*S[k,0:])
  A = mat_addcol(A,L[1],-A[0:,k])
  T = mat_addcol(T,L[1],-T[0:,k])
  A = mat_swaprow(A,L[0],k)
  A = mat_swapcol(A,L[1],k)
  S = mat_swaprow(S,L[0],k)
  T = mat_swapcol(T,L[1],k)
  return [S,A,T]


M1 = Mat([2,0],[0,3])
M2 = Mat([-4,-2,5],[3,5,7])
M3 = Mat([-4,3],[-2,5],[3,5])
M4 = Mat([5,6,8],[6,-11,7])
M5 = Mat([2,3],[3,4],[4,5],[6,7])
M6 = Mat([1,2,3],[4,5,6],[7,8,9])
M7 = Mat([2,0,0],[0,3,0],[0,0,0])
M8 = Mat([4,3,6,2],[2,3,6,4],[6,6,13,5])
M9 = Mat([1,0,0,0],[0,2,0,0],[0,0,3,0],[0,0,0,0])
M10 = Mat([0,2],[0,3])
M11 = Mat([0,0],[0,3])
M12 = Mat([0,0],[0,0])
M13 = Mat([0],[0])
M14 = Mat([-2,-1,-2,1],[5,3,4,-1],[1,0,1,-1],[-3,-1,-2,2]) # 斎藤 p.201
M15 = Mat([0])
M16 = Mat([-3])
ML = [M1,M2,M3,M4,M5,M6,M7,M8,M9,M10,M11,M12,M13,M14,M15,M16]

# for our test
def test1(*t):
  if len(t) and t[0]==1:
    for M in ML: print(elmdiv(M),"\n======="); input()
  else:
    for M in ML: print(elmdiv(M),"\n=======")


def elmdiv(M):
  n = min(M.shape)
  B = tab_gen(M)
  dprint("step0",-1,B)
  k = 0
  # don't replace B[1] with A, which behaves differently!
  while k < n and chk_zero(B[1],k)==False:
    X = step1(B,k)
    if X == False: break
    B = X
    dprint("step1",k,B)
    
    while True:
      while chk_col(B[1],k) == False:
        B = step2(B,k)
        dprint("step2",k,B)
        B = step1(B,k)
        dprint("step1",k,B)
      while chk_row(B[1],k) == False:
        B = step3(B,k)
        dprint("step3",k,B)
        B = step1(B,k)
        dprint("step1",k,B)
      if chk_col(B[1],k) == True: break
    
    # we may assume here both chk_col(B[1],k) and chk_row(B[1],k) are true
    if chk_mod(B[1],k): k += 1
    else:
      X = step4(B,k)
      if X==False: break
      B = X
      dprint("step4",k,B)
  r = chk_final(B,M)
  if r != 0 : error("# ERROR: bug!",r)
  return B
  
# B:elmdiv(M)



def lde_solution(B):
  S,A,T=B[0],B[1],B[2]
  b1=S.col(0)
  m=vlen(b1)
  n=min(A.shape[1],m)
  L=symbols(list(map(lambda x: "x%s"%x,range(1,A.shape[1]+1))))
  """-
  then L=[x1,x2,x3] if m==3
  range(A.shape[1]+1)))
  -"""
  for k in range(0,n):
    if A[k,k]==0:
      if b1[k] != 0: 
        L=None
        break
    else:
      dm = divmod(b1[k],A[k,k])
      if dm[1]==0: L[k]=dm[0]
      else:
        L=None
        break
  if L == None: return None
  # if some of b1[n],b1[n+1],..., are non-zero, we have no solution
  if len(sublist(b1[n:],lambda x: x != 0)) != 0: return None
  L = Mat(L).transpose()
  return T*L

# final check
def lde_final(L,M,b):
  c = Mat(b).transpose()
  if L==None: return False
  else:
    L1 = M*L
    if is_zero(expand(L1-c)): return True
    error("# ERROR: bug!",L,M,b,L1,c)

def lde(M,b):
  B = tab_gen2(M,b)
  if B == False: error("# ERROR,inconsistent b")
  n = min(M.shape)
  dprint("step0",-1,B)
  # don't replace B[1] below with A, which behaves differently!
  for k in range(0,n):
    if chk_zero(B[1],k)==True: break
    X = step1(B,k)
    if X == False: break
    B = X
    dprint("step1",k,B)
    while True:
      while chk_col(B[1],k)==False:
        B = step2(B,k)
        dprint("step2",k,B)
        B = step1(B,k)
        dprint("step1",k,B)
      while chk_row(B[1],k)==False:
        B = step3(B,k)
        dprint("step3",k,B)
        B = step1(B,k)
        dprint("step1",k,B)
      if chk_col(B[1],k)==True: break
  L = lde_solution(B)
  if lde_final(L,M,b) : return list(expand(L))
  else: return "no solution"

"""-
examples: 
M = Mat([5,6,8],[6,-11,7])
b=[1,9]
B = lde(M,b)
-"""
MD1 = [Mat([5,6,8],[6,-11,7]), [1,9]] # → [10*x3+5,x3,(-7*x3)-3]
MD2 = [Mat([5,6,8],[6,-11,7]), [1,8]] # → "no solution"
MD3 = [Mat([4,3,6,2],[2,3,6,4],[6,6,13,5]), [4,2,3]] # → [x4,10-4*x4,x4-4,x4-1]
MD4 = [Mat([0,0],[0,0],[0,2],[0,0]), [0,0,2,0]] # → [x2,1]
MD5 = [Mat([2,0],[0,0],[0,0],[0,0]), [2,1,0,0]] # → "no solution"
MD6 = [Mat([2,0],[0,0],[0,0],[0,0]), [2,0,1,0]] # → "no solution"
MDL = [MD1,MD2,MD3,MD4,MD5,MD6]

# for our test
# for our test
def test2(*t):
  if len(t) and t[0]==1:
    for MD in MDL: print(lde(MD[0],MD[1]),"\n======="); input()
  else:
    for MD in MDL: print(lde(MD[0],MD[1]),"\n=======")

def rand(n): return randint(0,n-1)

# mat_rand(n,m,nr): make matrix M of degree n randomly
# M is square, all the elements are integer, and the determinant is +1 or -1
# may be useful to find tutorial examples for your choice */
def mat_rand(n,m,nr):
  # n is degree, m is maximum multiplier, nr is repeat count
  M=eye(n)
  for i in range(0,nr):
    t = rand(3), # operation type
    n1 = rand(n)
    n2 = rand(n)
    print(n1,n2)
    while n1 == n2: n2 = rand(n)
    if rand(2)==0: # row
      if t == 0:
        M = mat_mulrow(M,n1,-1)
      elif t == 1:
        M = mat_swaprow(M,n1,n2)
      else:
        L = M.row(n1)
        M = mat_addrow(M,n2,L*(rand(m)-divmod(m,2)[0]))
    else: # col
      if t == 0:
        M = mat_mulcol(M,n1,-1)
      elif t == 1:
        M = mat_swapcol(M,n1,n2)
      else:
        L = M.col(n1)
        M = mat_addcol(M,n2,L*(rand(m)-divmod(m,2)[0]))
    print(i,M)
    if abs(det(M)) != 1: error("# ERROR: determinant")
  return(M)
# example: look the output for your choice 
# mat_rand(4,3,50);


